static char rcsid[] = "@(#)$Id: outheaders.c,v 1.2 2002/02/02 19:25:00 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.2 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *****************************************************************************/

#include "headers.h"
#include "hdr_imp.h"
#include "s_me.h"

DEBUG_VAR(Debug,__FILE__,"header");


static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}

void zero_mailing_headers (hdrs)
     struct mailing_headers *hdrs;
{
    /* bzero is defined hdrs/defs.h */
    bzero(hdrs,sizeof (*hdrs));
 
    hdrs->magic       = MAIL_HDR_magic;
    
    hdrs->subject     = NULL;
    zero_expanded_address(&(hdrs->from));
    zero_id_phrase(&(hdrs->in_reply_to));
    hdrs->expires     = NULL;
    hdrs->action      = NULL;
    hdrs->priority    = NULL;
    hdrs->sender    = NULL;
    zero_expanded_address(&(hdrs->reply_to));
    zero_expanded_address(&(hdrs->to));
    zero_expanded_address(&(hdrs->cc));
    hdrs->user_header       = NULL;
    hdrs->user_header_count = 0;
    zero_expanded_address(&(hdrs->bcc));
    hdrs->precedence   = NULL;
    hdrs->expires_days = NULL;
}

void free_mailing_headers (hdrs)
     struct mailing_headers *hdrs;
{
    if (hdrs->magic   != MAIL_HDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"free_mailing_headers",
	      "Bad magic number",0);

    if (hdrs->subject)
	free_string(&(hdrs->subject));
    hdrs->subject     = NULL;
    free_expanded_address(&(hdrs->from));
    free_id_phrase(&(hdrs->in_reply_to));
    if (hdrs->expires)
	free(hdrs->expires);
    hdrs->expires     = NULL;
    if (hdrs->action)
	free(hdrs->action);
    hdrs->action      = NULL;
    if (hdrs->priority)
	free(hdrs->priority);
    hdrs->priority    = NULL;
    if (hdrs->sender)
	free(hdrs->sender);
    hdrs->sender     = NULL;
    free_expanded_address(&(hdrs->reply_to));
    free_expanded_address(&(hdrs->to));
    free_expanded_address(&(hdrs->cc));

    if (hdrs->user_header) {
	int i;

	for (i = 0; i < hdrs->user_header_count; i++) {
	    if (hdrs->user_header[i].value)
		free_string(&(hdrs->user_header[i].value));
	    hdrs->user_header[i].name = NULL;
	}
	free(hdrs->user_header);
	hdrs->user_header       = NULL;
    }
    hdrs->user_header_count = 0;

    free_expanded_address(&(hdrs->bcc));
    if (hdrs->precedence)
	free(hdrs->precedence);
    hdrs->precedence   = NULL;
    if (hdrs->expires_days)
	free(hdrs->expires_days);
    hdrs->expires_days = NULL;

    /* Make sure that dangling pointters are catched */
    /* bzero is defined hdrs/defs.h */
    bzero(hdrs,sizeof (*hdrs));
    
}

int add_expanded_addr_ (x,ADR,FN,COM)
     struct expanded_address *x;
     CONST char *ADR;
     CONST struct string *FN;
     CONST struct string *COM;
{
    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"add_expanded_addr_",
	      "Bad magic number",0);

    x->addrs = safe_realloc(x->addrs,(x->addrs_len + 2) * 
			    sizeof (struct addr_item));
    x->addrs[x->addrs_len].addr     = safe_strdup(ADR);
    x->addrs[x->addrs_len].fullname = dup_string(FN);
    x->addrs[x->addrs_len].comment  = dup_string(COM);
    x->addrs[x->addrs_len+1].addr     = NULL;
    x->addrs[x->addrs_len+1].fullname = NULL;
    x->addrs[x->addrs_len+1].comment  = NULL;
    return x->addrs_len++;
}

int add_textual_addr_ (x,TEXTUAL,POS,LEN)
     struct expanded_address *x;
     CONST char *TEXTUAL;
     int POS, LEN;
{
    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"add_textual_addr_",
	      "Bad magic number",0);

    x->surface = safe_realloc(x->surface,(x->surface_len + 2) * 
			      sizeof (struct textual));
    if (!TEXTUAL)
	x->surface[x->surface_len].textual     = NULL;
    else
	x->surface[x->surface_len].textual     = safe_strdup(TEXTUAL);
    x->surface[x->surface_len].pos         = POS;
    x->surface[x->surface_len].len         = LEN;
    x->surface[x->surface_len+1].textual   = NULL;
    x->surface[x->surface_len+1].pos       = POS+LEN;
    x->surface[x->surface_len+1].len       = 0;
    return x->surface_len++;
}

void zero_expanded_address (x)
     struct expanded_address *x;
{
    /* bzero is defined hdrs/defs.h */
    bzero(x,sizeof (*x));
    x->magic        = EXP_ADDR_magic;

    x->addrs        = NULL;
    x->addrs_len    = 0;
    x->surface      = NULL;
    x->surface_len  = 0;
}

void copy_expanded_address (result,source) 
     struct expanded_address *result;
     struct expanded_address source;
{
    int i;
    free_expanded_address(result);

    if (source.magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"copy_expanded_address",
	      "Bad magic number",0);
    
    for (i = 0; i < source.surface_len; i++) {
	int count = 0;
	int pos = result->addrs_len;
	int j;
	for (j = 0; j < source.surface[i].len; j++) {
	    CONST struct addr_item *ptr1 = & 
		(source.addrs[source.surface[i].pos+j]);
	    int p = add_expanded_addr_(result,ptr1->addr,ptr1->fullname,
				       ptr1->comment);
	    if (0 == count)
		pos = p;
	    count++;
	}
	add_textual_addr_(result,source.surface[i].textual,pos,count);      
    }
}

void free_expanded_address(x)
     struct expanded_address *x;
{
    int i;
    
    if (x->magic   != EXP_ADDR_magic) 
    	panic("HEADERS PANIC",__FILE__,__LINE__,"free_expanded_address",
	      "Bad magic number",0);

    if (x->addrs_len)
	free_addr_items(x->addrs);
    x->addrs = NULL;
    x->addrs_len = 0;
    
    if (x->surface_len) {
	for (i = 0; i < x->surface_len; i++) {
	    if (x->surface[i].textual) {
		free(x->surface[i].textual);
		x->surface[i].textual = NULL;
	    }
	}
	free(x->surface);
	x->surface = NULL;
	x->surface_len = 0;
    }
}

char * make_surface_addr(ITEM)
     struct addr_item ITEM; 
{
    char * result = NULL;
    
    if (string_len(ITEM.fullname)) {
	struct string * temp = convert_string(display_charset,ITEM.fullname,1);
       	char * D = us2s(stream_from_string(temp,1,NULL)), *p;

	/* No fancy quotation stuff on textual form for editing */
	for (p = D; *p; p++) {	    	    
	    switch(*p) {
	    case '"': case '\\': 
		*p = '_';
		break;
	    }
	}

	result = strmcat(result,"\"");
	result = strmcat(result,D);
	result = strmcat(result,"\" ");
	free(D);
	free_string(&temp);
    }

    result = strmcat(result,"<");
    result = strmcat(result,ITEM.addr);
    result = strmcat(result,">");
    
    if (string_len(ITEM.comment)) {
	struct string * temp = convert_string(display_charset,ITEM.comment,1);
	char * D = us2s(stream_from_string(temp,1,NULL)), *p;

	/* No fancy quotation stuff on textual form for editing */
	for (p = D; *p; p++) {	    	    
	    switch(*p) {
	    case '"': case '\\': case '(': case ')':
		*p = '_';
		break;
	    }
	}

	result = strmcat(result," (");
	result = strmcat(result,D);
	result = strmcat(result,")");
	free(D);
	free_string(&temp);
		
    }

    return result;
}

#ifdef ANSI_C
hdr_add_to_mailing_hdr add_from_hdr;
#endif
int add_from_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     CONST char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    int j;

    struct addr_item * TMP;

    TMP = break_down_address(value,demime,defcharset);

    if (replace)
	free_expanded_address(&(headers->from));

    for (j = 0; TMP[j].addr && TMP[j].fullname; j++) {
	int pos = add_expanded_addr_(&(headers->from),
				     TMP[j].addr,
				     TMP[j].fullname,
				     TMP[j].comment);
	char *s = make_surface_addr(TMP[j]);
	add_textual_addr_(&(headers->from),s,pos,1);
	free(s);
    }

    free_addr_items(TMP);

    return 1;
}

#ifdef ANSI_C
hdr_add_to_mailing_hdr add_user_hdr;
#endif

int add_user_hdr(headers,X,value,demime,defcharset,replace)
     struct mailing_headers * headers;
     header_ptr X;
     const char *value;
     int demime;
     charset_t defcharset;
     int replace;
{
    if (replace) {
	int i;
	for (i = 0; i < headers->user_header_count; i++) {
	    if (headers->user_header[i].value)
		free_string(&(headers->user_header[i].value));
	    headers->user_header[i].name = NULL;
	}
	headers->user_header_count = 0;
    }

    headers->user_header = safe_realloc(headers->user_header,
				     (headers->user_header_count +1) *
				     sizeof (headers->user_header[0]));
    headers->user_header[headers->user_header_count].name = X;
    headers->user_header[headers->user_header_count].value = 
	hdr_to_string(HDR_TEXT,value,defcharset,demime);
    headers->user_header_count++;

    return 1;
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */

