static char rcsid[] = "@(#)$Id: mime_parse.c,v 1.34 2002/03/06 15:05:11 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.34 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 *
 *  Initially written by: Michael Elkins <elkins@aero.org>, 1995
 *****************************************************************************/

#include "headers.h"
#include "melib.h"
#include "s_me.h"

DEBUG_VAR(Debug,__FILE__,"mime");

void mime_t_clear (mt)
     mime_t *mt;
{
    DPRINT(Debug,15,(&Debug,
		     "mime_t_clear(%p) --> BEGIN\n",mt));

    if (mt->magic != MIME_magic)
	mime_panic(__FILE__,__LINE__,"mime_t_clear",
		   "Bad magic number");
    
    mt->offset = mt->begin_offset = 0;
    mt->length = -1;
    mt->encoding = ENCODING_7BIT;
    mt->unlink = 0;
    mt->TYPE        = give_media_type2(MIME_TYPE_TEXT,"plain",0);
    if (!mt->TYPE)
	mime_panic(__FILE__,__LINE__,"mime_t_clear",
		   "text/plain is not known");

    mt->disposition = DISP_INLINE;
    mt->notplain = 0;
    
    if (mt->parser_data)
	mime_parser_free(&(mt->parser_data));

#if 0
    if (mt->parts)
	mime_destroy (mt->parts);
    if (mt->next)
	mime_destroy (mt->next);

    mt->next = mt->prev = mt->parts = NULL;
#endif

    if (mt->description)
	free_string (&(mt->description));
    if (mt->type_opts)
	free (mt->type_opts);
    mt->type_opts = NULL;
    if (mt->disposition_opts)
	free (mt->disposition_opts);
    mt->disposition_opts = NULL;
    
    if (mt->pathname)
	free (mt->pathname);
    mt->pathname = NULL;
    if (mt->dispname)
	free_string(& (mt->dispname));
        
    DPRINT(Debug,15,(&Debug,
		     "mime_t_clear(%p) <-- END\n",mt));
    return;
}

void
mime_get_disposition (str, mt)
     char *str;
     mime_t *mt;
{
    char *c, tmp[VERY_LONG_STRING];
    
    DPRINT(Debug,9,(&Debug,
		    "mime_get_disposition(): str=\"%s\"\n", str));
    
    if (mt->magic != MIME_magic)
	mime_panic(__FILE__,__LINE__,"mime_get_disposition",
		   "Bad magic number");
    
    /* Don't harm "str" */
    strfcpy (tmp, str, sizeof(tmp));
    
    rfc822_reap_comments (tmp, NULL, 0);
    
    /* Look for the options field */
    if ((c = strchr (tmp, ';')) != NULL) {
	char *d = c;
	while (d > tmp && whitespace(*(d-1)))
	    d--;
	*d = '\0';
	c++;
	while (*c && whitespace(*c))
	    c++;
	mt->disposition_opts = strmcpy (mt->disposition_opts, c);
    }
    else {
	char *d = tmp + strlen(tmp);
	while (d > tmp && whitespace(*(d-1)))
	    d--;
	*d = '\0';
	if (mt->disposition_opts) {
	    free (mt->disposition_opts);
	    mt->disposition_opts = NULL;
	}
    }
    
    /* All that's left now is the main disposition */
    c = tmp;
    while (*c && whitespace(*c))
	c++;
    /* No Content-Disposition -header     -> DISP_INLINE     (DISP_AUTOATTACH)
     *    Content-Disposition: inline     -> DISP_INLINE
     *    Content-Disposition: attachment -> DISP_ATTACH
     *    Content-Disposition: {unknown}  -> DISP_ATTACH
     * See RFC 1806 (Experimental protocol) for details.
     */   
    if (istrcmp (c, "inline") != 0)
	mt->disposition = DISP_ATTACH;
    else
	mt->disposition = DISP_INLINE;
    
    DPRINT(Debug,9,(&Debug,
		    "mime_get_disposition(): disposition=\"%s\" (%d), disposition_opts=\"%s\"\n",
		    DISPOSITION(mt->disposition), 
		    mt->disposition,
		    NONULL(mt->disposition_opts)));
}

void
mime_get_content (str, mt)
     char *str;
     mime_t *mt;
{
    char *c, tmp[VERY_LONG_STRING];

    char * subtype = "???";
    
    DPRINT(Debug,9,(&Debug,
		    "mime_get_content(): str=\"%s\"\n", str));
    
    if (mt->magic != MIME_magic)
	mime_panic(__FILE__,__LINE__,"mime_get_content",
		   "Bad magic number");
    
    /* Don't harm "str" */
    strfcpy (tmp, str, sizeof(tmp));
    
    rfc822_reap_comments (tmp, NULL, 0);
    
    /* Look for the options field */
    if ((c = strchr (tmp, ';')) != NULL) {
	char *d = c;
	while (d > tmp && whitespace(*(d-1)))
	    d--;
	*d = '\0';
	c++;
	while (*c && whitespace(*c))
	    c++;
	mt->type_opts = strmcpy (mt->type_opts, c);
    }
    else {
	char *d = tmp + strlen(tmp);
	while (d > tmp && whitespace(*(d-1)))
	    d--;
	*d = '\0';
	if (mt->type_opts) {
	    free (mt->type_opts);
	    mt->type_opts = NULL;
	}
    }
        
    /* Get the subtype */
    if ((c = strchr (tmp, '/')) != NULL) {
	char *d = c;
	while (d > tmp && whitespace(*(d-1)))
	    d--;
	*d = '\0';
	
	c++;
	while (*c && whitespace(*c))
	    c++;
	subtype = c;
    }
    
    /* All that's left now is the main type */
    c = tmp;
    while (*c && whitespace(*c))
	c++;

    mt->TYPE = give_media_type(c,subtype,1);
    
    DPRINT(Debug,9,(&Debug,
		    "mime_get_content(): %p -- type=\"%s\", subtype=\"%s\", opts=\"%s\"\n",
		    mt->TYPE,
		    get_major_type_name(mt->TYPE), 
		    get_subtype_name(mt->TYPE), 
		    NONULL(mt->type_opts)));
    
  return;
}

int mime_get_boundary (boundary, opts, size)
     char *opts, *boundary;
     int size;
{
    if (!mime_get_param ("boundary", boundary, opts, size)) {
	lib_error(CATGETS(elm_msg_cat, MeSet, MeParseNoBoundary,
			  "'boundary' paramater is missing from Multipart -type!"));
	return 0;
    }

    return 1;
}

int class_charset(charset_value) 
     char * charset_value;
{
    /* Return > 0  (2)  if charset is displayable with display_charset
     * Return < 0  (-2) if charset needs translating to display_charset
     */
    int ret = 0;
    charset_t charset_value_p = MIME_name_to_charset(charset_value,0);
    
    /* Check agaist display charset */
    if (0 == istrcmp(charset_value,"US-ASCII") && 
	charset_ok_p(display_charset))
	ret = 2; 
    /* If display_charset can show us-ascii? */
    else if (display_charset->MIME_name &&
	     0 == istrcmp(charset_value,display_charset->MIME_name))
	ret = 2;
    /* if charset is subset of display */
    else if (charset_value_p && charset_superset_of(display_charset,
						    charset_value_p))
	ret = 2;
    else if (charset_value_p && 
	     (CS_printable & charset_properties(charset_value_p)))
	ret = -2;  /* convert (filter?) to display charset */
    else
	ret = 0;
    
    DPRINT(Debug,9,(&Debug,
		    "class_charset: charset_value=\"%s\"\n",
		    charset_value));
    DPRINT(Debug,9,(&Debug,
		    "             : display_charset=\"%s\"\n",
		    display_charset->MIME_name ? 
		    display_charset->MIME_name :
	      "<no MIME name>"));
    DPRINT(Debug,9,(&Debug,
		    "class_charset=%d\n",ret));

    return ret;
}

int mime_get_charset (charset_value, opts, display_charset,
		      default_content_charset)
     CONST char *opts; 
     charset_t *charset_value;
     charset_t * display_charset; /* vector */
     charset_t default_content_charset;
{  
    /* Return > 0  (2)  if charset is displayable with display_charset     
     * Return < 0  (-2) if charset needs translating to display_charset
                    -3  if charset can be translated to display_charset
                        without loss
     */
    int ret = 0,j;
    char buffer[80];

    if (!mime_get_param("charset",buffer,opts,sizeof buffer)) {
	if (default_content_charset && 
	    default_content_charset->MIME_name) {
	    DPRINT(Debug,9,(&Debug,
			    "mime_get_charset: Using content default charset %s\n",
			    default_content_charset->MIME_name));
	    *charset_value = default_content_charset;
			    
	} else {
	    strfcpy(buffer,"US-ASCII", sizeof buffer); 
	    /* Default charset if nothing specified */
	    
	    *charset_value = MIME_name_to_charset(buffer,CHARSET_create);
	}
    } else
	*charset_value = MIME_name_to_charset(buffer,CHARSET_create);

    if (!opts) {
	DPRINT(Debug,9,(&Debug,
			"mime_get_charset: opts=NULL\n"));
    } else {
	DPRINT(Debug,9,(&Debug,
			"mime_get_charset: opts=\"%s\"\n",opts));
    }
    DPRINT(Debug,9,(&Debug,
		    "                : charset_value=%p '%s' (type=%p)\n",
		    *charset_value,
		    (*charset_value)->MIME_name,
		    (*charset_value)->charset_type));
    
    for (j = 0; display_charset[j]; j++) {
	DPRINT(Debug,9,(&Debug,
			"            [%d] : display_charset=%p '%s' (type=%p)\n",
			j,display_charset[j],
			display_charset[j]->MIME_name ? 
			display_charset[j]->MIME_name :
			"<no MIME name>",
			display_charset[j]->charset_type));
	
	
	if (display_charset[j]->MIME_name &&
	    0 == istrcmp(buffer,display_charset[j]->MIME_name)) {
	    ret = 2; /* If display charset              */
	    break;
	} else if (charset_superset_of(display_charset[j],*charset_value)) {
	    ret = 2; /* if charset is subset of display */
	    break;
	} else if (0 == istrcmp(buffer,"US-ASCII") && 
		   charset_ok_p(display_charset[j])) {
	    ret = 2; /* If display_charset can show us-ascii? */
	    break;
	}
    }
    if ( 0 == ret) {
	int prop;
	if (CS_printable & charset_properties(*charset_value)) {
	    ret = -2;  /* convert (filter?) to display charset */
	    if (display_charset[0] &&
		CS_universal_set & (prop = 
				    charset_properties(display_charset[0]))
		) {
		DPRINT(Debug,9,(&Debug,
				"          display charset '%s' --  universal charset (flags=%d)\n",
				display_charset[0]->MIME_name ?
				display_charset[0]->MIME_name :
				"<no MIME name>",
				prop));
		ret = -3;  /* convert to display charset without loss */
	    }
	} else
	    ret = 0;
    }
    DPRINT(Debug,9,(&Debug,
		    "mime_get_charset=%d\n",ret));

    return ret;
}

void
mime_t_zero (ptr)
     mime_t *ptr;
{
  /* This routine should be called whenever a new "mime_t" is created.  It
   * makes sure that the pointers inside are correctly initialized to NULL
   * so that we don't end up calling free() on an uninitialized pointer.
   */
    DPRINT(Debug,15,(&Debug,
		     "mime_t_zero(%p)\n",ptr));

    ptr->parser_data = NULL;
#if 0
  ptr->next = ptr->parts = NULL;
#endif
  ptr->type_opts = ptr->disposition_opts = 
      ptr->pathname = NULL;
  ptr->dispname = NULL;
  ptr->description = NULL;

  ptr->magic = MIME_magic;
  

  ptr->offset = ptr->begin_offset = 0;
  ptr->length = -1;
  ptr->encoding = ENCODING_7BIT;
  ptr->unlink = 0;
  ptr->TYPE        = give_media_type2(MIME_TYPE_TEXT,"plain",0);
  if (!ptr->TYPE)
      mime_panic(__FILE__,__LINE__,"mime_t_zero",
		 "text/plain is not known");

  ptr->disposition = DISP_INLINE;
  ptr->notplain = 0;
}

void mime_t_copy(trg, src)
     mime_t *trg, *src;
{
  /* This routines make copy of mime_t structure ... */

    DPRINT(Debug,15,(&Debug,
		     "mime_t_copy(%p,%p) --> BEGIN\n",trg,src));
  
  if (trg->magic != MIME_magic)
    mime_panic(__FILE__,__LINE__,"mime_t_copy",
	       "Bad magic number (trg)");

  if (src->magic != MIME_magic)
    mime_panic(__FILE__,__LINE__,"mime_t_copy",
	       "Bad magic number (src)");

  mime_t_clear(trg);

  trg->offset =            src->offset; 
  trg->begin_offset  =     src->begin_offset; 
  trg->length =            src->length;
  trg->encoding =          src->encoding;
  trg->unlink = 0;         /* Don't unlink ! */
  trg->TYPE =              src->TYPE;
  trg->disposition =       src->disposition;
  trg->notplain =          src->notplain;


  trg->parser_data = NULL;        
  copy_parser_data(& (trg->parser_data), src->parser_data);

  if (src->description) {
      trg->description = dup_string(src->description);
  }

  if (src->type_opts) {
      trg->type_opts = strmcpy(trg->type_opts, src->type_opts);
  }

  if (src->disposition_opts) {
      trg->disposition_opts = strmcpy(trg->disposition_opts, 
				      src->disposition_opts);
  }

  DPRINT(Debug,15,(&Debug,
		   "mime_t_copy(%p,%p) <-- END\n",trg,src));
}

mime_t *
mime_t_alloc ()
{
  mime_t *ptr;

  DPRINT(Debug,15,(&Debug,
		   "mime_t_alloc()     --> BEGIN\n"));

  ptr = (mime_t *) safe_malloc (sizeof (mime_t));
  /* Make sure to clear the pointers initially so that later we know when
   * to reclaim memory in mime_t_clear().
   */
  mime_t_zero (ptr);
  mime_t_clear (ptr);

    DPRINT(Debug,15,(&Debug,
		     "mime_t_alloc() = %p <-- END\n",ptr));
    return ptr;
}

void parse_mime_headers1 (ptr,headers,part_offset,body_offset,opts, 
			  hdr_charset)
     mime_t *ptr;
     header_list_ptr headers;
     long part_offset;
     long body_offset;
     int opts;
     charset_t hdr_charset;
{
    header_list_ptr this_header;
    int no_content_disposition = 0;

    /* set some defaults */
    ptr->encoding = ENCODING_7BIT;
    if (opts & MIME_DIGEST) {
	ptr->TYPE = give_media_type2(MIME_TYPE_MESSAGE,"rfc822",0);
	if (!ptr->TYPE)
	    mime_panic(__FILE__,__LINE__,"parse_mime_headers1",
		       "message/rfc822 is not known");
    } else {
	ptr->TYPE = give_media_type2(MIME_TYPE_TEXT,"plain", 0);
	if (!ptr->TYPE)
	    mime_panic(__FILE__,__LINE__,"parse_mime_headers1",
		       "text/plain is not known");
    }

    ptr->disposition = DISP_INLINE;
    ptr->description = NULL;
    
    ptr->begin_offset = part_offset;
    ptr->offset       = body_offset;
    ptr->length = -1;
    
    if (NULL != (this_header = 
		 locate_header_by_name(headers,"Content-Type")) &&
	NULL != this_header->body) {
	mime_get_content (this_header->body, ptr);
	if (this_header->next_this_header &&
	    show_header_errors) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeParseErrorContent,
			"PARSE ERROR: Several Content-Type headers!"));
	}
    } else {
	DPRINT(Debug,9,(&Debug,
			"parse_mime_headers1: No Content-Type -header\n"));
    }

    if (NULL != (this_header = 
		 locate_header_by_name(headers,"Content-Disposition")) &&
	NULL != this_header->body) {
	mime_get_disposition (this_header->body, ptr);
	if (this_header->next_this_header &&
	    show_header_errors) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeParseErrorDisposition,
			      "PARSE ERROR: Several Content-Disposition headers!"));
	}
    } else {
	DPRINT(Debug,9,(&Debug,
			"parse_mime_headers1: No Content-Disposition -header\n"));
	no_content_disposition = 1;
    }

    if (NULL != (this_header = 
		 locate_header_by_name(headers,"Content-Transfer-Encoding")) &&
	NULL != this_header->body) {
	char  * value = this_header->body, *c;
	
	/* This removes comments from buffer this_header->body */
	
	rfc822_reap_comments (value, NULL, 0);
	c = value;
	while (*c && isspace((unsigned char) *c))
	    c++;
	ptr->encoding = check_encoding (c);
	
	if (this_header->next_this_header &&
	    show_header_errors) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeParseErrorTransfer,
			      "PARSE ERROR: Several Content-Transfer-Encoding headers!"));
	}
    } else {
	DPRINT(Debug,9,(&Debug,
			"parse_mime_headers1: No Content-Transfer-Encoding -header\n"));
    }
    
    if (NULL != (this_header = 
		 locate_header_by_name(headers,"Content-Description"))  &&
	NULL != this_header->body) {
	char * c = this_header->body;

	while (*c && whitespace(*c))
	    c++;
	ptr->description = NULL;
	ptr->description = hdr_to_string(HDR_TEXT,c,
					 hdr_charset,
					 1);
					 	
	if (this_header->next_this_header &&
	    show_header_errors) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeParseErrorDescription,
			      "PARSE ERROR: Several Content-Description headers!"));
	}
    } else {
	DPRINT(Debug,9,(&Debug,
			"parse_mime_headers1: No Content-Description -header\n"));
    }
    
    if (mime_notplain(ptr)) {
	ptr->notplain = TRUE;
	DPRINT(Debug,9,(&Debug,
			"parse_mime_headers1: 'Not plain'\n"));
	if (no_content_disposition) {
	    DPRINT(Debug,9,(&Debug,
			    "                   : no content-disposition -- auto-attachment = %d\n",
			    auto_attachment));
	    if (get_major_type_code(ptr->TYPE) == MIME_TYPE_APPLICATION && 
		auto_attachment >= 1 ||
		
		get_major_type_code(ptr->TYPE) != MIME_TYPE_TEXT       && 
		!(get_type_flags(ptr->TYPE) & MIME_RFC822)             &&
	        get_major_type_code(ptr->TYPE) != MIME_TYPE_MULTIPART  && 
		auto_attachment >= 2) {
		DPRINT(Debug,9,(&Debug,
				"                   : Set disposition = auto attachment\n"));
		ptr->disposition = DISP_AUTOATTACH;
	    }		
	}
    }    
}

mime_t * parse_mime_headers(headers,part_offset,body_offset,opts,
			    hdr_charset)
     header_list_ptr headers;
     long part_offset;
     long body_offset;
     int opts;
     charset_t hdr_charset;
{
  mime_t *ptr;

  DPRINT(Debug,9,(&Debug,
		  "parse_mime_headers(): part_offset=%ld, body_offset=%ld, opts=%d\n",
		  part_offset,body_offset,opts));

  ptr = mime_t_alloc ();

  parse_mime_headers1(ptr,headers,part_offset,body_offset,opts,hdr_charset);


  DPRINT(Debug,10,(&Debug,
		   "parse_mime_headers- type=%s/%s; flags=%d;\n",
		   get_major_type_name(ptr->TYPE), 
		   get_subtype_name(ptr->TYPE), 
		   get_type_flags(ptr->TYPE)));
  DPRINT(Debug,10,(&Debug,
		   "parse_mime_headers- begin=%ld, offset=%ld, length=%ld\n",
		   ptr->begin_offset,ptr->offset,ptr->length));

  DPRINT(Debug,9,(&Debug,
		  "parse_mime_headers=%p <-- END\n",(void *)ptr));

  return ptr;
}

mime_t * mime_read_header (fp, opts, defcharset)
     FILE *fp;
     int opts;
     charset_t defcharset;
{
  mime_t *ptr;
  header_list_ptr headers = NULL;
  long part_offset;
  long body_offset;


  DPRINT(Debug,19,(&Debug,
		   "mime_read_header: opts=%d --> START\n",opts));

  part_offset = ftell (fp);
  headers = file_read_headers(fp,0);
  body_offset = ftell(fp);

  ptr = parse_mime_headers(headers,part_offset,body_offset,opts,
			   defcharset);

  delete_headers(&headers);

  return ptr;
}
 
void
mime_warnings(hdr) 
     struct header_rec *hdr;
{

  if (hdr->status & PRE_MIME_CONTENT) {
    lib_error(CATGETS(elm_msg_cat, MeSet, MeParsePreMime,
		      "Error: MIME-message has pre-MIME content-type!"));
  }
  if (hdr->status & MIME_UNSUPPORTED) {
    lib_error(CATGETS(elm_msg_cat, MeSet, MeParseUnsupportedMime,
		      "Warning: Unsupported MIME-Version!"));
  }
}

#if ANSI_C
parse_mime_callback mime_parse_routine;
#endif

int mime_parse_routine (folder,hdr, fp)
     struct folder_info *folder;
     struct header_rec *hdr;
     FILE *fp;
{
    /* This routine checks to see if the multipart messages specified by
     * "hdr" has been parsed for its subparts, and if not, calls the routine
     * to do so.
     */
    
    if (hdr->mime_rec.magic != MIME_magic)
	mime_panic(__FILE__,__LINE__,"mime_parse_routine",
		   "Bad magic number (mime_rec)");
    
    if (hdr->mime_rec.parser_data || !(hdr->status & MIME_MESSAGE))
	return 1;

    /* Copy value */
    if (hdr -> content_length >= 0)
	hdr->mime_rec.length = hdr -> content_length;
    
    if (hdr->mime_rec.begin_offset <= 0) {
	
	fseek(fp,hdr->offset,SEEK_SET);
	hdr->mime_rec.begin_offset = hdr->offset;
	/* Skip mailbox's separator lines ... */
	
	DPRINT(Debug,9,(&Debug,
			"mime_parse_routine: scanning begin_offset: %ld\n",
			(long) hdr->mime_rec.begin_offset));
	
	hdr->mime_rec.begin_offset = skip_envelope(hdr,fp);
	
	DPRINT(Debug,9,(&Debug,
			"mime_parse_routine: begin_offset=%ld\n",
			(long) hdr->mime_rec.begin_offset));
	
	if (hdr->mime_rec.begin_offset < 0) {
	    lib_error(CATGETS(elm_msg_cat, MeSet, MeParseMailError,
			      "Can't parse mail..."));
	    return 0;
	}
    }

    lib_transient(CATGETS(elm_msg_cat, MeSet, MeParsingMime,
			  "Parsing MIME structure..."));
    if (mime_parser_parse(&(hdr->mime_rec),hdr->header_charset,fp)) {
	int tmp = mime_notplain(&(hdr->mime_rec));
	if (tmp != hdr->mime_rec.notplain) {
	    hdr->mime_rec.notplain = tmp;
	    if (!tmp) {
		lib_error(CATGETS(elm_msg_cat, MeSet, MeParsingMimeNoMetamail,
				  "Parsing MIME structure... metamail not needed"));	    
	    }
	}
	hdr->mime_parsed            = 1;
    }

    return 1;
}

#if 0
int is_text_type (primary_type, subtype, enc)
     char *primary_type, *subtype;
     int enc;
{
    /* encoding rules are different for text and no text 
     * for text we must do \n -> \r\n before base64 encoding */
    
    /* Current MIME Draft Standard says that part is text
     * (ie. line oriented) when it is Text/ -type -- otherwise no.
     *
     * new draft for Mime standards allows also other parts ot be text.
     * It says that content-transfer-encodings: 7bit and 8bit are only
     * allowed for line orienteed types */
    
    /* Actually Message/ and Multipart/ can't be newer encoded directly */
    if ( istrcmp(primary_type,"Message") == 0 ||
	 istrcmp(primary_type,"Multipart") == 0)
	return -1;
    
    if (istrcmp(primary_type,"Text") == 0) 
	return 1;               /* Subtypes of text are always text (ie.
				 * line oriented). */
    
    
    if (enc == ENCODING_NONE || enc == ENCODING_7BIT || enc == ENCODING_8BIT)
	return 1;             /* It is text */
    
    if (enc == ENCODING_BINARY)
	return 0;            /* It is probably binary (or very long lines) */
  
    return 0;              /* Default: It is binary */
}
#endif

int is_pre_mime_content_type (ptr,content_type)
     mime_t *ptr;
     char *content_type;
{
    int result;
    char *cptr = strpbrk(content_type,"/;()");

    DPRINT(Debug,10,(&Debug,
		     "is_pre_mime_content_type(): content_type=%s\n",
		     content_type));
    
    if (!cptr || ';' == *cptr) {
	char *ptr2;
	char *tmp = content_type;
	while (whitespace(*tmp))
	    tmp++;
	
	if (cptr)
	    *cptr = '\0';
	
	ptr2 = strpbrk(tmp," \t");
	if (ptr2)
	    *ptr2 = '\0';
	
	if (istrcmp(tmp,"text")!=0) {
	    char buf[STRING];
	    ptr -> notplain = TRUE;
	    
	    /* Put some 'intelligent' value */
	    elm_sfprintf(buf,sizeof buf,
			 FRM("X-RFC1049-%.30s"),tmp);
	    ptr->TYPE = give_media_type2(MIME_TYPE_APPLICATION,
					 buf,1);
	} else {
	    ptr -> notplain = FALSE;
	    ptr -> TYPE = give_media_type2(MIME_TYPE_TEXT,"plain",0);
	    if (!ptr->TYPE)
		mime_panic(__FILE__,__LINE__,"is_pre_mime_content_type",
			   "text/plain is not known");
 
	}
	result = 1;
    } else 
	result = 0;
    
    DPRINT(Debug,10,(&Debug,
		     "is_pre_mime_content_type=%d\n",result));
    
    return result;
}


/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
