static char rcsid[] = "@(#)$Id: addr_util.c,v 1.28 2002/02/02 19:25:01 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.28 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 * 			Copyright (c) 1988-1992 USENET Community Trust
 * 			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** This file contains addressing utilities 

**/

#include "headers.h"
#include "s_elm.h"
#include "me.h"

DEBUG_VAR(Debug,__FILE__,"addr");

static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}
static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}


int aliases_to_expanded(x)
     struct expanded_address *x;
{
    int i;
    int tagged = 0;
    CONST int NEWPOS = x->addrs_len;
    
    for (i=0; i < message_count; i++) {
	if (ison(aliases[i]->status, TAGGED)) {
	    add_textual_addr_(x,aliases[i]->alias,NEWPOS,0);
	    tagged++;
	}
    }
    
    if (tagged == 0) {
	add_textual_addr_(x,aliases[current-1]->alias,NEWPOS,0);
    }

    return tagged;
}

void dump_expanded_address(debuglevel,text,expanded)
     int debuglevel; 
     CONST char *text;
     struct expanded_address expanded;

{
    struct textual *ptr;
    int l = strlen(text);
    int LASTPOS = -1;

    DPRINT(Debug,debuglevel,
	   (&Debug, 
	    "%s: %d surfaces, %d addresses\n",
	    text, expanded.surface_len, expanded.addrs_len));
 
    for (ptr = expanded.surface; 
	 ptr < expanded.surface + expanded.surface_len; 
	 ptr++) {
	int j;
	if (!ptr->textual) {
	    DPRINT(Debug,debuglevel,
		   (&Debug, 
		    "%*s [%d] surface NULL! \t(%d addresses)\n",
		    l,"",ptr-expanded.surface,ptr->len));
	} else {
	    DPRINT(Debug,debuglevel,
		   (&Debug, 
		    "%*s [%d] surface=%s \t(%d addresses)\n",
		    l,"",ptr-expanded.surface,ptr->textual,
		    ptr->len));
	}
	if (ptr->pos != LASTPOS + 1) {     
	    DPRINT(Debug,debuglevel,
		   (&Debug, 
		    "%*s [%d] ERROR!  address pos = %d != %d (lastpost +1)\n",
		    l,"",ptr-expanded.surface, ptr->pos, LASTPOS+1));			
	}
	for (j = 0; j < ptr->len; j++) {
	    if (ptr->pos < 0 || ptr->pos >= expanded.addrs_len) {
		DPRINT(Debug,debuglevel,
		       (&Debug, 
			"%*s    %d [%d] ERROR, not range 0 - %d\n",
			l,"",j,ptr->pos,
			expanded.addrs_len-1));
	    } else {
		CONST struct addr_item *ptr1 = & (expanded.addrs[ptr->pos]);
		if (!ptr1->addr || !ptr1->fullname || !ptr1->comment) {
		    DPRINT(Debug,debuglevel,
			   (&Debug, 
			    "%*s    %d [%d] ERROR! NULL element!\n",
			    l,"",j,ptr->pos));
		} else {
		    DPRINT(Debug,debuglevel,
			   (&Debug, 			    
			    "%*s    %d [%d] addr=%s \tfullname=%S\tcomment=%S\n",
			    l,"",j,ptr->pos,
			    ptr1->addr,ptr1->fullname,
			    ptr1->comment));
		}
		LASTPOS = ptr->pos;
	    }
	}
    }
    DPRINT(Debug,debuglevel,
	   (&Debug, 
	    "%*s === end\n",l,""));
}

extern char **argv_from_headers (headers)
     struct mailing_headers * headers;
{
    int count = 
	headers->to.addrs_len + 
	headers->cc.addrs_len + headers->bcc.addrs_len;
    char **res;
    int idx = 0,i;
    struct addr_item *p;
        
    DPRINT(Debug,6,
	   (&Debug,  "argv_from_headers, count=%d\n",count));
    
    res = safe_malloc((count + 1) * sizeof (char *));
    
    dump_expanded_address(6,"argv_from_headers -- enter (to)",headers->to);

    for (p = headers->to.addrs; 
	 p < headers->to.addrs + headers->to.addrs_len;
	 p++) {
	if (p->addr[0] == '-' || p->addr[0] == '@' ||
	    p->addr[0] == '\0')
	    /* Modify p->addr so that there is space
	     * alloced also for <addr> form 
	     */
	    res[idx++] = kludge_addr(&p->addr);
	else
	    res[idx++] = p->addr;
    }

    dump_expanded_address(6,"argv_from_headers -- enter (cc)",headers->cc);

    for (p = headers->cc.addrs; 
	 p < headers->cc.addrs + headers->cc.addrs_len;
	 p++) {
	if (p->addr[0] == '-' || p->addr[0] == '@' ||
	    p->addr[0] == '\0')
	    /* Modify p->addrs so that there is space
	     * alloced also for <addr> form 
	     */
	    res[idx++] = kludge_addr(&p->addr);
	else
	    res[idx++] = p->addr;
    }

    dump_expanded_address(6,"argv_from_headers -- enter (bcc)",headers->bcc);

    for (p = headers->bcc.addrs; 
	 p < headers->bcc.addrs + headers->bcc.addrs_len;
	 p++) {
	if (p->addr[0] == '-' || p->addr[0] == '@' ||
	    p->addr[0] == '\0')
	    /* Modify p->addrs so that there is space
	     * alloced also for <addr> form 
	     */
	    res[idx++] = kludge_addr(&p->addr);
	else
	    res[idx++] = p->addr;
    }
    
    res[idx] = NULL;
    
    DPRINT(Debug,6,
	   (&Debug, "           idx=%d\n",idx));
    for (i = 0; i < idx; i++) 
	DPRINT(Debug,6,
	       (&Debug, "           [%d]=%s\n",i,res[i]));

    dump_expanded_address(7,"argv_from_headers -- leave (to)",headers->to);
    dump_expanded_address(7,"argv_from_headers -- leave (cc)",headers->cc);
    dump_expanded_address(7,"argv_from_headers -- leave (bcc)",headers->bcc);

    return res;
}


int argv_to_expanded(result,argv,mailer_info)
     struct expanded_address *result;
     char *argv[];
     struct mailer_info      *mailer_info;
{
    int i,res;
    free_expanded_address(result);

    DPRINT(Debug,6,
	   (&Debug, "argv_to_expanded:"));
    for (i = 0;  argv[i]; i++) {
	DPRINT(Debug,6,
	       (&Debug, " [%d]=%s",i,argv[i]));
    }
    DPRINT(Debug,6,
	   (&Debug,"\n"));

    for (i = 0;  argv[i]; i++) {
	add_textual_addr_(result,argv[i],0,0);
    }

    dump_expanded_address(6,"argv_to_expanded: before build_address_l()",
			  *result);

    /* And make address structures */
    res = build_address_l(result,mailer_info);

    dump_expanded_address(6,"argv_to_expanded: (result)",*result);

    DPRINT(Debug,6,
	   (&Debug, "argv_to_expanded=%d\n",res));
    return res;
}

void addr_to_expanded(result,addrs,mailer_info)
     struct expanded_address *result;
     struct addr_item *addrs;
     struct mailer_info      *mailer_info;
{
    int i;
    free_expanded_address(result);

    DPRINT(Debug,6,
	   (&Debug ,"addr_to_expanded:"));
    if (!addrs) {
	DPRINT(Debug,6,
	       (&Debug ," -- NULL"));
    } else {
	for (i = 0;  
	     addrs[i].addr && addrs[i].fullname && addrs[i].comment; i++) {
	    DPRINT(Debug,6,
		   (&Debug, 		    
		    " [%d]: {addr=%s, fullname=%S, comment=%S}",
		    i,
		    addrs[i].addr, addrs[i].fullname, 
		    addrs[i].comment));
	    
	}
    }
    DPRINT(Debug,6,
	   (&Debug, "\n"));
    
    for (i = 0;  
	 addrs && addrs[i].addr && addrs[i].fullname && addrs[i].comment; 
	 i++) {
	int pos = add_expanded_addr_(result,addrs[i].addr, 
				addrs[i].fullname, addrs[i].comment);
	add_textual_addr_(result,NULL,pos,1);
    }

    dump_expanded_address(7,"addr_to_expanded: before build_address_l()",
			  *result);

    /* And make address structures */
    build_address_l(result,mailer_info);

    dump_expanded_address(6,"addr_to_expanded: (result)",*result);
}

void expanded_to_edit_buffer(buffer,size,expanded)
     char * buffer;
     int size;  
     struct expanded_address expanded;
{
    struct textual *ptr;
    buffer[0] = '\0';

    DPRINT(Debug,6,
	   (&Debug, "expanded_to_edit_buffer: size=%d\n",size));
    dump_expanded_address(6,"expanded_to_edit_buffer",expanded);

    for (ptr = expanded.surface; 
	 ptr < expanded.surface + expanded.surface_len; 
	 ptr++) {
	if (buffer[0]) 
	    strfcat(buffer,", ",size);
	strfcat(buffer,ptr->textual,size);
    }
    DPRINT(Debug,6,
	   (&Debug, "expanded_to_edit_buffer: (result) buffer=%s\n",
	    buffer));
}

void update_expanded_from_edit_buffer(expanded,buffer,mailer_info)
     struct expanded_address *expanded;
     CONST char *buffer;
     struct mailer_info      *mailer_info;
{
    int i,next;
    struct expanded_address result;
    struct textual *ptr = expanded->surface;
    char ** tokenized = rfc822_tokenize(buffer);

    zero_expanded_address(&result);

    DPRINT(Debug,6,
	   (&Debug, "update_expanded_from_edit_buffer: buffer=%s\n",
	    buffer));
    dump_expanded_address(6,"expanded_to_edit_buffer: (initial)",*expanded);

    for (i = 0; tokenized[i]; i = next) {
	CONST int NEWPOS = result.addrs_len;
	char * surface = NULL;
	int q= 0, spacecount =  0, q_seen = 0;

	DPRINT(Debug,9,
	       (&Debug, 
		"expanded_to_edit_buffer: [%d]=\"%s\"\n",
		i,tokenized[i]));

	if (whitespace(tokenized[i][0]) ||
	    ',' == tokenized[i][0]) {
	    next = i+1;
	    continue;
	}
	for (next = i; tokenized[next]; next++) {
	    if ('<' == tokenized[next][0]) {
		q++;
		q_seen = 1;
	    } else if ('>' == tokenized[next][0])
		q--;
	    else if (!q && ',' == tokenized[next][0])
		break;
	    else if (!q && q_seen && 
		     !whitespace(tokenized[next][0]) &&
		     '(' != tokenized[next][0])
		break;
	    else if (!q && whitespace(tokenized[next][0]) &&
		     tokenized[next+1] && '(' == tokenized[next+1][0])
		/* Don't count whitespaces before comments */  ;
	    else if (!q && q_seen &&
		     whitespace(tokenized[i][0]))
		break;
	    else if (!q && whitespace(tokenized[next][0]))
		spacecount ++;

	    DPRINT(Debug,9,
		   (&Debug, 
		    "expanded_to_edit_buffer+ [%d]=\"%s\" q=%d\n",
		    next,tokenized[next],q));
	    
	    surface = strmcat(surface,tokenized[next]);     	       
	}

	DPRINT(Debug,9,
	       (&Debug, 
		"expanded_to_edit_buffer- spacecount=%d, q=seen=%d\n",
		spacecount,q_seen));

  
	if (surface && ptr < expanded->surface + expanded->surface_len &&
	    0 == strcmp(ptr->textual,surface)) {
	    int r;
	    int pos = NEWPOS;
	    int count = 0;
	    int j;
	    /* no change -- copy original */
      
	    for (j = 0; j < ptr->len; j++) {
		CONST struct addr_item *ptr1 = & (expanded->addrs[ptr->pos+j]);
		int p = ADD_EXPANDED(result,ptr1->addr,ptr1->fullname,
				     ptr1->comment);
		if (0 == count)
		    pos = p;
		count++;
	    }
	    r=ADD_TEXTUAL(result,ptr->textual,pos,count);
	    DPRINT(Debug,9,
		   (&Debug, 
		    "expanded_to_edit_buffer-> [%d] = %s (keep %d addresses)\n",
		    r,surface,count));
	    
	}

	/* Starting with - is alias expansion removing syntax */
	else if (tokenized[i][0] == '-') {
	    int r;
	    char * surface1 = NULL;
      
	    DPRINT(Debug,9,
		   (&Debug, 
		    "expanded_to_edit_buffer= resscan from %d\n",i));
			   
	    for (next = i; tokenized[next]; next++) {
		if ('<' == tokenized[next][0] ||
		    '>' == tokenized[next][0] ||
		    ',' == tokenized[next][0] ||
		    whitespace(tokenized[next][0]))
		    break;

		DPRINT(Debug,9,
		       (&Debug, 
			"expanded_to_edit_buffer+ [%d]=\"%s\"\n",
			next,tokenized[next]));
		
		surface1 = strmcat(surface1,tokenized[next]);     	       
	    }
	    
	    r=ADD_TEXTUAL(result,surface1,NEWPOS,0);
	    DPRINT(Debug,9,
		   (&Debug, 
		    "expanded_to_edit_buffer-> [%d] = %s\n",
		    r,surface1));
	    
	    if (surface1)
		free(surface1);
	    
	    /* word <address> */
	} else if (spacecount <= 1 && q_seen) {
	    int r = ADD_TEXTUAL(result,surface,NEWPOS,0);
	    DPRINT(Debug,9,
		   (&Debug, "expanded_to_edit_buffer-> [%d] = %s\n",
		    r,surface));
	} else {
	    char * surface1 = NULL;
	    int r;

	    DPRINT(Debug,9,
		   (&Debug, 
		    "expanded_to_edit_buffer= resscan from %d\n",i));
	    
	    for (next = i; tokenized[next]; next++) {
		if (!q && whitespace(tokenized[next][0]) &&
		    tokenized[next+1] && '(' == tokenized[next+1][0])
		    /* Don't break on whitespaces before comments */  ;
		else if ('<' == tokenized[next][0] ||
			 '>' == tokenized[next][0] ||
			 ',' == tokenized[next][0] ||
			 whitespace(tokenized[next][0]))
		    break;

		DPRINT(Debug,9,
		       (&Debug, 
			"expanded_to_edit_buffer+ [%d]=\"%s\"\n",
			next,tokenized[next]));

		surface1 = strmcat(surface1,tokenized[next]);     	       
	    }

	    r=ADD_TEXTUAL(result,surface1,NEWPOS,0);
	    DPRINT(Debug,9,
		   (&Debug, "expanded_to_edit_buffer-> [%d] = %s\n",
		    r,surface1));
	    
	    if (surface1)
		free(surface1);
	}

	if (ptr < expanded->surface + expanded->surface_len)
	    ptr++;
	if (surface)
	    free(surface);


	if (i == next) {
	    lib_error(CATGETS(elm_msg_cat, ElmSet, ElmUnparseable,
			      "Unparseable (%s): %s"),
		      tokenized[next],buffer);
	    next++;
	}
    }

    dump_expanded_address(6,"expanded_to_edit_buffer: before build_address_l",
			  result);

    build_address_l(&result,mailer_info);

    free_expanded_address(expanded);
    *expanded = result;   /* put new (same?) result to place */

    if (tokenized)
	free(tokenized);

    dump_expanded_address(6,"expanded_to_edit_buffer: (result)",*expanded);
}

char * from_addr_literal(mailer_info)
     struct mailer_info *mailer_info;
{
    char * Q = NULL;

    /* If on username is strange characters (specially if username
       is from some other user database backend than from /etc/passwd)
       then quote it...
    */
    
    if (NULL != strpbrk(username,
			":\"\\!#%&()=?',;.:<>")) {
	Q = elm_message(FRM("%Q"),username);

	DPRINT(Debug,6,
	       (&Debug, "from_addr_literal: Quoting username: %s\n",Q));

    } else 
	Q = safe_strdup(username);


    if (mailer_info &&
	query_mailer_info(mailer_info,MI_USE_DOMAIN)) {
	Q = strmcat(Q,"@");
	Q = strmcat(Q, hostfullname);
    }

    return Q;
}

int make_from_addr(expanded,mailer_info)
     struct expanded_address *expanded;
     struct mailer_info *mailer_info;
{
    int res;
    int pos;


    /* FIXME: Posisble wrong charset */
    struct string * comment = new_string(display_charset);
    struct string * name    = new_string2(display_charset,s2us(full_username));

    char * Q = from_addr_literal(mailer_info);

    free_expanded_address(expanded);

    pos = add_expanded_addr_(expanded,Q,name,comment);
    free(Q); Q = NULL;
   
    add_textual_addr_(expanded,NULL,pos,1);

    dump_expanded_address(6,"make_from_addr: before build_address_l",
			  *expanded);

    /* And make address structures */
    res = build_address_l(expanded,mailer_info);

    dump_expanded_address(6,"make_from_addr: (result)",*expanded);
        DPRINT(Debug,6,
	       (&Debug, "make_from_addr=%d\n",res));

    free_string(&name);
    free_string(&comment);

    return res;
}

int build_address_l(expanded,mailer_info)
    struct expanded_address *expanded;
    struct mailer_info      *mailer_info;
{
    struct textual *ptr;
    int expands = 0;
    struct expanded_address result;
    int too_long = FALSE;

    dump_expanded_address(9,"build_address_l",*expanded);

    zero_expanded_address(&result);

    for (ptr = expanded->surface; 
	 ptr < expanded->surface + expanded->surface_len; 
	 ptr++) {
	CONST int NEWPOS = result.addrs_len;
	struct addr_item * aliases;

	struct addr_item verify_result;

	if (!ptr->textual) {

	    if (ptr->len == 1) {
		/* If we have no textual presentation we are converting
		 * from struct addrs
		 */
		char *temp = make_surface_addr(expanded->addrs[ptr->pos]);
		int pos = ADD_EXPANDED(result,
				       expanded->addrs[ptr->pos].addr,
				       expanded->addrs[ptr->pos].fullname,
				       expanded->addrs[ptr->pos].comment);
		ADD_TEXTUAL(result,temp,pos,1);
	
		free(temp);
	    } else {
		DPRINT(Debug,1,
		       (&Debug, 
			"build_address_l: SOFWARE ERROR: [%d]: textual=NULL, len=%d\n",
			ptr - expanded->surface, ptr->len));
	    }
	}

	else if (ptr->textual[0] == '-') {
	    /* Is also on result's remove list */
	    ADD_TEXTUAL(result,ptr->textual,NEWPOS,0);
	}

	else if (qstrpbrk(ptr->textual,"!@: <>,()") != NULL) {
	    char * temp = NULL;
	    if (ptr->len == 1) 
		temp = make_surface_addr(expanded->addrs[ptr->pos]);

	    if (temp && 0 == strcmp(temp,ptr->textual)) {
		/* no change -- preserve original address */

		int pos = ADD_EXPANDED(result,
				       expanded->addrs[ptr->pos].addr,
				       expanded->addrs[ptr->pos].fullname,
				       expanded->addrs[ptr->pos].comment);
		ADD_TEXTUAL(result,temp,pos,1);

	    } else {
		/* generate addresses from textual 'surface' presentation */
		int pos = NEWPOS;
		int count = 0;
		struct addr_item * address =
		    break_down_address(ptr->textual, 
				       /* If user pastes encoded words from 
					* somewhere decode them also.
					*/
				       is_rfc1522(ptr->textual),
				       display_charset);
				 
		struct addr_item *ptr1;

		for (ptr1 = address; 
		     ptr1 && ptr1 -> addr && ptr1->fullname && ptr1->comment; 
		     ptr1++) {
		    int p;
		    
		    if (qstrpbrk(ptr1->addr,"!:@") == NULL) {
			if (mailer_info &&
			    query_mailer_info(mailer_info,MI_USE_DOMAIN)) {
			    ptr1->addr = strmcat(ptr1->addr,"@");
			    ptr1->addr = strmcat(ptr1->addr, hostfullname);
			}
		    }

		    p = ADD_EXPANDED(result,ptr1->addr,
				      ptr1->fullname, ptr1->comment);
		    if (0 == count)
			pos = p;
		    count++;
		}
		ADD_TEXTUAL(result,ptr->textual,pos,count);
		if (address)
		    free_addr_items(address);
	    }

	    if (temp)
		free(temp);

	} 
	
	else if (NULL != (aliases = 
			  get_alias_address_l(ptr->textual, TRUE, 
					      &too_long,mailer_info))) {
	    struct addr_item *ptr1;
	    int pos = NEWPOS;
	    int count = 0;

	    expands = 1;

	    for (ptr1 = aliases; 
		 ptr1 && ptr1 -> addr && ptr1->fullname && ptr1->comment; 
		 ptr1++) {
		struct textual *ptr2;
		int p;

		/* scan words to be eleminated */
		for (ptr2 = expanded->surface; 
		     ptr2 < expanded->surface + expanded->surface_len; 
		     ptr2++) {
		    if (ptr2 -> textual &&
			ptr2 -> textual[0] == '-') {
			if (0 == strcmp(ptr1->addr,&(ptr2->textual[1])))
			    break;
		    }    
		}

		if (ptr2 && ptr2 < expanded->surface + expanded->surface_len)
		    /* Is on eliminating list -- don't add */
		    continue;

		p = ADD_EXPANDED(result,ptr1->addr,ptr1->fullname,
				 ptr1->comment);
		if (0 == count)
		    pos = p;
		count++;
	    }
	    ADD_TEXTUAL(result,ptr->textual,pos,count);

	    if (aliases)
		free_addr_items(aliases);
	} 

	else if (too_long) {
	    DPRINT(Debug,2,
		   (&Debug, "Overflowed alias expansion for %s\n", 
		    ptr->textual));
	    continue;             /* Don't process */
	}

	else if (mailer_info &&
		 verify_mailer_addr(mailer_info,ptr->textual,
				    &verify_result)) {
	    int pos;

	    DPRINT(Debug,5,
		   (&Debug, 
		    "Address %s verified -- really %s\n",
		    ptr->textual,verify_result.addr));
		    		    
	    /* Preserve possible comment and fullname
	     *
	     * Use dup_string() to avoid shared pointer so
	     * there no need to make spacial case for free_string()
	     */

	    if (ptr->len == 1 && ! verify_result.fullname)
		verify_result.fullname = 
		    dup_string(expanded->addrs[ptr->pos].fullname);
	    if (ptr->len == 1 && ! verify_result.comment)
		verify_result.comment = 
		    dup_string(expanded->addrs[ptr->pos].comment);

	    if (!verify_result.fullname)
		verify_result.fullname = new_string(display_charset);
	    if (!verify_result.comment)
		verify_result.comment = new_string(display_charset);
	       
	    pos = ADD_EXPANDED(result,verify_result.addr,
			       verify_result.fullname,
			       verify_result.comment);
 	    ADD_TEXTUAL(result,ptr->textual,pos,1);

	    /* ADD_EXPANDED makes copy, so original need to be free'ed 
	     * on above there is make duplicate from 
	     *   expanded->addrs[ptr->pos].{fullname,comment) so 
	     * this can be free'ed safely	     	    
	     */
	    free_string(&verify_result.fullname);
	    free_string(&verify_result.comment);
	    free(verify_result.addr); verify_result.addr = NULL;
	}

	else { 
	    struct string * gecos  = NULL; 
	    struct string * comment = NULL;
	    int pos; 

	    /* Preserve possible comment 
	     *
	     * Use dup_string() to avoid shared pointer so
	     * there no need to make spacial case for free_string()
	     */
	    if (ptr->len == 1)
		gecos = dup_string(expanded->addrs[ptr->pos].fullname);
	    if (ptr->len == 1)
		comment = dup_string(expanded->addrs[ptr->pos].comment);

	    if (!gecos)
		gecos = new_string(display_charset);
	    if (!comment)
		comment = new_string(display_charset);

	    pos = ADD_EXPANDED(result,ptr->textual,gecos,comment);
	    if (mailer_info &&
		query_mailer_info(mailer_info,MI_USE_DOMAIN)) {
		result.addrs[pos].addr = strmcat(result.addrs[pos].addr,"@");
		result.addrs[pos].addr = strmcat(result.addrs[pos].addr, 
						 hostfullname);
	    }
	    ADD_TEXTUAL(result,ptr->textual,pos,1);
	    
	    if (check_only) 
		lib_error(CATGETS(elm_msg_cat, ElmSet, ElmAliasUnknown,
				  "(alias \"%s\" is unknown)"), 
			  ptr->textual);

	    /* ADD_EXPANDED makes copy, so original need to be free'ed
	     * on above there is make duplicate from 
	     *   expanded->addrs[ptr->pos].{fullname,comment) so 
	     * this can be free'ed safely
	     */
	    free_string(&gecos);
	    free_string(&comment);
	}
    }

    free_expanded_address(expanded);
    *expanded = result;   /* put new (same?) result to place */

    dump_expanded_address(9,"build_address_l: (result)",*expanded);
    DPRINT(Debug,9,
	   (&Debug,"build_address_l=%d\n",expands));
    
    return expands;
}

void write_addr_header(mailer,hdr_name,addr,top_encoding,enmime,hdr_charset)
     out_state_t *mailer;
     char *hdr_name;
     struct addr_item * addr;
     int top_encoding;
     int enmime;
     charset_t hdr_charset;
{
    struct addr_item *p;
    long POS = out_state_ftell(mailer);

#define LEN (out_state_ftell(mailer)-POS)
#define WRAP print_EOLN(mailer,top_encoding), \
 (POS = out_state_ftell(mailer)), state_putc(' ',mailer)

    state_puts(hdr_name,mailer);
    state_puts(": ",mailer);

    DPRINT(Debug,9,
	   (&Debug, "write_addr_header: (enmime=%d) %s:",
	    enmime,hdr_name));
    
    for (p = addr; p && p->fullname && p->addr && p->comment; p++) {
	CONST int al = strlen(p->addr);
	CONST int fn = string_len(p->fullname);
	CONST int cm = string_len(p->comment);

	DPRINT(Debug,9,
	       (&Debug, "\n          [%d] ",p-addr));
	
	if (p != addr) {
	    state_puts(", ",mailer);
	    DPRINT(Debug,9,
		   (&Debug, ", "));
	}
	if (LEN > 70 || al + fn + cm + LEN > 76) {
	    WRAP;
	    DPRINT(Debug,9,
		   (&Debug,"[WRAP]\n          [%d] ",p-addr));
	}
	
	if (fn || ! p->addr[0] || '@' == p->addr[0]) {
	    if (fn) {
		char * buffer = string_to_hdr(HDR_PHRASE, p->fullname,
					      hdr_charset, enmime);
		char * c;

		DPRINT(Debug,9,
		       (&Debug,"{fullname quoted=%s}\n",
			buffer));
		DPRINT(Debug,9,
		       (&Debug, "\n          [%d] ",p-addr));
	       		
		for (c = buffer; *c != '\0'; c++) {
		    if (*c == '\n' || 
			(whitespace(*c) && (LEN > 120))) {
			WRAP;
			DPRINT(Debug,9,
			       (&Debug, "[WRAP]\n          [%d] ",
				p-addr));
		    } else {
			state_putc(*c,mailer);
			DPRINT(Debug,9,
			       (&Debug, "%c",*c));
		    }
		}
		free(buffer);
	    }

	    if (LEN > 120 || (LEN > 70 && fn > 70)) {
		WRAP;
		DPRINT(Debug,9,
		       (&Debug, "[WRAP]\n          [%d] ",p-addr));
	    } else {
		state_putc(' ',mailer);
		DPRINT(Debug,9,
		       (&Debug ," "));
	    }

	    state_putc('<',mailer);
	    state_puts(p->addr,mailer);
	    state_putc('>',mailer);
	    DPRINT(Debug,9,
		   (&Debug, "<%s>",p->addr));
	} else {
	    state_puts(p->addr,mailer);
	    DPRINT(Debug,9,
		   (&Debug ,"%s",p->addr));
	}
	if (cm) {
	    char * buffer = string_to_hdr(HDR_COMMENT, p->comment,
					  hdr_charset, enmime);
	    char * c;
	    
	    DPRINT(Debug,9,
		   (&Debug,"\n          [%d] ",p-addr));

	    if (LEN > 85) {
		DPRINT(Debug,9,
		       (&Debug, "[WRAP]\n          [%d] ",p-addr));
		WRAP;
	    } else {
		state_putc(' ',mailer);
		DPRINT(Debug,9,
		       (&Debug ," "));
	    }
	    
	    DPRINT(Debug,9,
		   (&Debug, "\n          {comment encoded=%s}",buffer));
	    DPRINT(Debug,9,
		   (&Debug, "\n          [%d] ",p-addr));

	    state_putc('(',mailer);			
	    DPRINT(Debug,9,
		   (&Debug, "("));
	    for (c = buffer; *c != '\0'; c++) {
		if (*c == '\n' || 
		    (whitespace(*c) && (LEN > 100))) {
		    WRAP;
		    DPRINT(Debug,9,
			   (&Debug, "[WRAP]\n          [%d] ",
			    p-addr));
		} else {
		    state_putc(*c,mailer);
		    DPRINT(Debug,9,
			   (&Debug, "%c",*c));
		}
	    }
	    state_putc(')',mailer);			
	    DPRINT(Debug,9,
		   (&Debug, ")"));
	    free(buffer);
	}
    }
    print_EOLN(mailer,top_encoding);
    DPRINT(Debug,9,
	   (&Debug, "[EOLN]\n"));
}

int check_8bit_addr(addr)
     struct addr_item * addr;
{
    struct addr_item *p;
    for (p = addr; p && p->fullname && p->addr && p->comment; p++) {
	if (!can_ascii_string(p->fullname))
	    return TRUE;
	if (!can_ascii_string(p->comment))
	    return TRUE;
    }
    return FALSE;
}

void write_text_header (mailer,hdr_name,text, top_encoding)
     out_state_t *mailer;
     char *hdr_name; 
     char * text;
     int top_encoding;
{
    long POS = out_state_ftell(mailer);
    char *c;

    state_puts(hdr_name,mailer);
    state_puts(": ",mailer);

    DPRINT(Debug,9,
	   (&Debug,"write_text_header: %s:",hdr_name));
    
    for (c = text; *c != '\0'; c++) {
	if (*c == '\n' || 
	    (whitespace(*c) && (LEN > 75))) {
	    WRAP;
	    DPRINT(Debug,9,
		   (&Debug,"[WRAP]\n          "));
	} else {
	    state_putc(*c,mailer);
	    DPRINT(Debug,9,
		   (&Debug,"%c",*c));
	}
    }
    print_EOLN(mailer,top_encoding);    
    DPRINT(Debug,9,
	   (&Debug,"[EOLN]\n"));
}

#undef LEN
#undef WRAP

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
