static char rcsid[] = "@(#)$Id: mkhdrs.c,v 1.19 2002/02/02 19:25:01 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 1.19 $   $State: Exp $
 *
 *  Modified by: Kari Hurtta <hurtta+elm@ozone.FMI.FI>
 ******************************************************************************
 *  The Elm Mail System 
 *
 *			Copyright (c) 1988-1992 USENET Community Trust
 *			Copyright (c) 1986,1987 Dave Taylor
 *****************************************************************************/

/** This contains all the header generating routines for the ELM
    program.

**/

#include "headers.h"
#include "me.h"
#include "s_elm.h"

#include <errno.h>
extern int errno;  

DEBUG_VAR(Debug,__FILE__,"header");

char *elm_date_str();


static unsigned char *s2us P_((char *str));
static unsigned char *s2us(str) 
     char *str;
{
    return (unsigned char *)str;
}


void generate_reply_to(current_header,headers)
     struct header_rec * current_header;
     struct mailing_headers *headers;
{
    /** Generate an 'in-reply-to' message... **/
    char date_buf[SLEN];
    int found = 0;

    free_id_phrase(&(headers->in_reply_to));

    if (add_irt_phrase) {
	headers->in_reply_to.text = new_string(display_charset);
    
	if (current_header->messageid) {
	    headers->in_reply_to.id = safe_strdup(current_header->messageid);
	    add_ascii_to_string(headers->in_reply_to.text,s2us(" from "));
	} else
	    add_ascii_to_string(headers->in_reply_to.text,s2us("From "));
	
	if (current_header) {	
	    struct addr_item *p;
	    
	    for (p = current_header->from; p && p->addr && p->fullname; p++) {
		
		if (found)
		    add_ascii_to_string(headers->in_reply_to.text,s2us(", "));
		
		found++;
		
		if (string_len(p->fullname)) {
		    struct string * tmp = 
			cat_strings(headers->in_reply_to.text,
				    p->fullname,1);
		    free_string(&(headers->in_reply_to.text));
		    headers->in_reply_to.text = tmp;		
		}
		else
		    add_ascii_to_string(headers->in_reply_to.text,
					s2us(p->addr));	    
	    }
	    
	    if (!found && current_header->env_from[0]) {
		add_ascii_to_string(headers->in_reply_to.text,s2us("(env: "));
		add_ascii_to_string(headers->in_reply_to.text,
				    s2us(current_header->env_from));
		add_ascii_to_string(headers->in_reply_to.text,s2us(")"));
	    } else if (!found)
		add_ascii_to_string(headers->in_reply_to.text,s2us("someone"));


	    add_ascii_to_string(headers->in_reply_to.text,s2us(" at "));
	    add_ascii_to_string(headers->in_reply_to.text,
				s2us(elm_date_str(date_buf,
						  current_header->time_sent +
						  current_header->tz_offset,
						  sizeof date_buf)));
	}


    } else if (current_header->messageid) {
	/* RFC 2822 (obsoletes RFC 822) says that in-reply-to
	   header should only include message-id
	*/
	headers->in_reply_to.id = safe_strdup(current_header->messageid);
    }
}

/* Returns either backquote replaced string or original
   if return value != buffer, value is malloced 
*/

static char * eval_backquote P_((char *buffer));
static char * eval_backquote(buffer)
     char *buffer;
{
    char *ptr ;
    char *result  = NULL;
    char *x1      = buffer;
    char *cptr    = NULL;
    int quitflag  = 0;

    DPRINT(Debug,8,(&Debug, "eval_backquote: %s\n",
		    buffer));


    for (ptr = buffer; (*ptr || result) && !quitflag; ptr++) {

	if ('`' == *ptr || !*ptr) {
	    
	    char * start = cptr ? cptr : x1;
	    int len;
	    char * clip;

	    quitflag = !*ptr;

	    if (!start)
		panic("HEADERS PANIC",__FILE__,__LINE__,"eval_backquote",
		      "No start pointer",0);
	    
	    len = ptr - start;
	    if (len < 0)
		panic("HEADERS PANIC",__FILE__,__LINE__,"eval_backquote",
		      "Bad start pointer",0);
		
	    clip = safe_malloc(len+1);
	    memcpy(clip,start,len);
	    clip[len] = '\0';

	    DPRINT(Debug,11,(&Debug, "eval_backquote: clip(len=%d)=%s\n",
			     len,clip));

	    if (!cptr) {
		/* Got start of command */

		result = strmcat(result,clip);
		cptr = ptr +1;
		x1 = NULL;
		free(clip);
		continue;
	    }

	    if (len > 0) {
		struct run_state RS;
		CONST char * argv[4];
		int retcode;

		char local_buffer[80];
		int n;

		int res;

		DPRINT(Debug,5,(&Debug, "eval_backquote: command: %s\n",
				clip));

		argv[0] = shell;
		argv[1] = "-c";
		argv[2] = clip;
		argv[3] = NULL;
		
		res = start_run(&RS,
				SY_ENAB_SIGHUP|SY_ENAB_SIGINT|
				SY_RUN_STATE_OPIPE,argv,-1,-1);
		if (0 == res) {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmBackquoteCmdFailed,
				      "Backquoted command \"%s\" in elmheaders failed."),
			      clip);
		    free(clip);
		    free(result);

		    DPRINT(Debug,5,(&Debug, 
				    "eval_backquote: start_run failed\n"));
		    return NULL;
		}

		while (0 < (n = 
			    mail_gets(local_buffer, sizeof local_buffer, 
				      RS.pfd))) {	
		    DPRINT(Debug,5,(&Debug, 
				    "eval_backquote: output: %s",
				    local_buffer));
		    if (local_buffer[n-1] != '\n') {
			DPRINT(Debug,5,(&Debug, 
				    "<-- no NEWLINE\n"));
		    }

		    result = strmcat(result,local_buffer);
		    
		}

		if (result && 0 < (n = strlen(result)) &&
		    '\n' == result[n-1]) {
		    result[n-1] = '\0';
		    DPRINT(Debug,5,(&Debug, 
				    "... Removed newline from end of result\n"));
		}

		res = wait_end(&RS,&retcode);
		if (res < 0 || retcode != 0) {
		    lib_error(CATGETS(elm_msg_cat, ElmSet, 
				      ElmBackquoteCmdFailed,
				      "Backquoted command \"%s\" in elmheaders failed."),
			      clip);
		    free(clip);
		    free(result);
		    
		    DPRINT(Debug,5,(&Debug, 
				    "eval_backquote: command failed: res=%d, retcode=%d\n",
				    res,retcode));
		    return NULL;

		}
	    }

	    x1 = ptr +1;
	    cptr = NULL;
	    free(clip);
	}
    }

    if (result) {
	DPRINT(Debug,8,(&Debug, "eval_backquote: result=%s\n",
			result));
    }
	
    return result ? result : buffer;
}



void import_mailheaders(headers,mime_info)
     struct mailing_headers * headers;
     mime_send_t * mime_info;
{
    int size;
    int err;
    FILE *F;
    char buffer[32*1024];

    if (0 != (err = can_open(user_mailheaders,"r"))) {
	DPRINT(Debug,5,(&Debug, "%s not available: %s (errno=%d)\n",
			user_mailheaders,error_description(err),err));
	return;
    }

    F = fopen(user_mailheaders,"r");
    if (!F) {
	/* In than point opening should have succeed */
	int err = errno;
	lib_error(FRM("%s: %s"),user_mailheaders,err);
	
	return;
    }

    while (0 < (size = read_header_line(F,buffer,sizeof buffer,0))) {

	header_ptr header_name;

	char *k;
	char * value1 = NULL;

	value1 = eval_backquote(buffer);

	if (!value1)
	    continue;

	k = strchr(value1,':');
	if (!k)
	    break;
	*k = '\0';
	k++;
	
	while (whitespace(*k))
	    k++;
	
	header_name = find_header(value1,1);
	
	if (!add_to_mailing_header(headers,header_name,k,
				   !allow_no_hdrencoding,
				   system_charset,0)) {
	    /* TODO: Notice failure ... */
	}

	if (value1 != buffer)
	    free(value1);

    }

    fclose(F);
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 * End:
 */
